from .single_table_operators import polygonize, explode_multi_geometry, spatial_aggregation, AggregationFunction, AggregationType
from .multi_table_operators import spatial_join, JoinType

__all__ = [
    "polygonize",
    "explode_multi_geometry",
    "spatial_aggregation",
    "AggregationFunction",
    "AggregationType",
    "spatial_join",
    "JoinType"
]



from enum import Enum

from .commons import DataFrame
from .single_table_operators import spatial_aggregation, AggregationFunction
import pyspark.sql as ps
import pyspark.sql.functions as func

from libadalina_core.sedona_utils import to_spark_dataframe
from sedona.sql import ST_Intersects

class JoinType(Enum):
    INNER = 'inner'
    LEFT = 'left'
    RIGHT = 'right'
    FULL = 'full'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

def spatial_join(
        left_table: DataFrame,
        right_table: DataFrame,
        join_type: JoinType = JoinType.INNER,
        aggregate: bool = False,
        aggregate_functions: list[AggregationFunction] | None = None
    ) -> ps.DataFrame:

    left_table = to_spark_dataframe(left_table)
    right_table = to_spark_dataframe(right_table)

    result = (left_table
              .withColumnRenamed('geometry', 'geometry_left')
              .join(right_table.withColumnRenamed('geometry', 'geometry_right'),
                    on=ST_Intersects(func.col('geometry_left'), func.col('geometry_right')), how=join_type.value)
              )

    if aggregate:
        if aggregate_functions is None:
            raise ValueError("aggregate_functions must be provided when aggregate is True")
        result = spatial_aggregation(result, aggregate_functions)

    return result




from libadalina_core.readers.geopackage import geopackage_to_dataframe
import pathlib
import pandas as pd

from libadalina_core.spatial_operators.multi_table_operators import spatial_join, JoinType, spatial_aggregation, AggregationType, \
    AggregationFunction

if __name__ == "__main__":
    """Example of how to use libadalina to find hospitals in specific provinces in Italy and aggregate their data."""

    # Set pandas display options
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', 100)

    hospitals = geopackage_to_dataframe(
        str(pathlib.Path(__file__).parent.parent / "tests" / "samples" / "healthcare" / "EU_healthcare.gpkg"),
        "EU"
    )[["hospital_name", "geometry", "city", "cap_beds"]]

    regions = geopackage_to_dataframe(
        str(pathlib.Path(__file__).parent.parent / "tests" / "samples" / "regions" / "NUTS_RG_20M_2024_4326.gpkg"),
        "NUTS_RG_20M_2024_4326.gpkg"
    )[["LEVL_CODE", "NUTS_NAME", "CNTR_CODE", "geometry"]]

    # select province of Milan and Cremona
    filtered_regions = regions[
        (regions['LEVL_CODE'] == 3) &
        (regions['CNTR_CODE'] == "IT") &
        (regions['NUTS_NAME'].str.contains('Milano|Cremona', case=False))
    ]

    # join with hospitals table to get hospitals in these provinces
    result = (spatial_join(filtered_regions, hospitals, join_type=JoinType.LEFT)
              # join operator renames the geometries adding suffixes _left and _right to avoid conflicts
              .withColumnRenamed('geometry_left', 'geometry'))
    result.show(truncate=False)

    # get the number of hospitals in each province along with the total and average number of beds
    result = spatial_aggregation(result, aggregate_functions=[
        AggregationFunction("hospital_name", AggregationType.COUNT, 'hospitals'),
        AggregationFunction("cap_beds", AggregationType.SUM, 'total_beds'),
        AggregationFunction("cap_beds", AggregationType.AVG, 'average_beds'),
    ])
    result.show(truncate=False)




from pyspark.sql import SparkSession
from sedona.spark import SedonaContext
import pandas as pd
from libadalina_core.sedona_configuration.jdk_installer import install_jdk_if_needed

# compatibility with Pandas
pd.DataFrame.iteritems = pd.DataFrame.items

def _get_sedona_master_configuration(master_host: str) -> SparkSession:
    spark = (SparkSession.builder
            .appName("Adalina")
            .master(master_host)
            .config(
                "spark.jars.packages",
                "org.apache.sedona:sedona-spark-3.3_2.12:1.7.1,"
                "org.datasyslab:geotools-wrapper:1.7.1-28.5",
            )
            .config(
                "spark.jars.repositories",
                "https://artifacts.unidata.ucar.edu/repository/unidata-all"
            )
            .config("spark.executor.instances", 1)
            .config("spark.executor.cores", "1")
            .config("spark.executor.memory", "2G")
            .getOrCreate())
    return SedonaContext.create(spark)

def _sedona_configuration() -> SparkSession:
    config = (
        SedonaContext.builder()
             .appName("Adalina")
             .config(
                "spark.jars.packages",
                "org.apache.sedona:sedona-spark-3.3_2.12:1.7.1,"
                "org.datasyslab:geotools-wrapper:1.7.1-28.5",
            )
            .config("spark.driver.memory", "20g")
            .config(
                "spark.jars.repositories",
                "https://artifacts.unidata.ucar.edu/repository/unidata-all"
            )
            .getOrCreate()
    )
    return SedonaContext.create(config)

_sedona_context: SparkSession | None = None

def init_sedona_context(
        spark_master: str | None = None,
        spark: SparkSession | None = None
):
    """
    Initialize the Sedona context for spatial data processing.

    This function can either:

    1. Create a new Sedona context with a specified Spark master,
    2. Use an existing SparkSession, or
    3. Create a default Sedona context with the default Spark configuration.

    If no parameters are provided, it will create a default Sedona context (option 3).

    If a `JAVA_HOME` environment variable is not set, it will attempt to install a compatible JDK.

    :param spark_master: The Spark master URL to connect to. If provided, a new Sedona context will be created with this master.
    :param spark: An existing SparkSession to use. If provided, it will be used to create the Sedona context.

    Example:
        # Initialize with default configuration
        init_sedona_context()

        # Initialize with specific Spark master
        init_sedona_context(spark_master="spark://localhost:7077")

        # Initialize with existing SparkSession
        spark = SparkSession.builder.getOrCreate()
        init_sedona_context(spark=spark)
    """
    global _sedona_context

    install_jdk_if_needed()

    if spark_master is not None:
        _sedona_context = _get_sedona_master_configuration(spark_master)
    elif isinstance(spark, SparkSession):
        _sedona_context = SedonaContext.create(spark)
    else:
        _sedona_context = _sedona_configuration()

def get_sedona_context() -> SparkSession:
    """
    Get the Sedona context for spatial data processing.
    This context is the one used for all spatial operations in libadalina.

    If the Sedona context has not been initialized yet with `init_sedona_context`,
    the function `init_sedona_context` will be called to initialize it with the default configuration.

    :return: The Sedona context as a SparkSession.
    """
    global _sedona_context

    if _sedona_context is None:
        init_sedona_context()
    return _sedona_context



import pandas as pd
import geopandas as gpd
import pyspark.sql as ps
from libadalina_core.sedona_utils import DEFAULT_EPSG

def dataframe_to_geopackage(df: pd.DataFrame | gpd.GeoDataFrame | ps.DataFrame, path: str):
    """
    Write a DataFrame to a GeoPackage file.
    DataFrame geometry is assumed to be in libadalina default EPSG `DEFAULT_EPSG`

    :param df: The DataFrame to write, which can be a pandas DataFrame, a GeoPandas GeoDataFrame, or a Spark DataFrame.
    :param path: The path to the GeoPackage file where the DataFrame will be saved.
    """
    if isinstance(df, ps.DataFrame):
        df = gpd.GeoDataFrame(df.toPandas(), geometry = 'geometry', crs = DEFAULT_EPSG.value)
    elif isinstance(df,  pd.DataFrame):
        df = gpd.GeoDataFrame(df, geometry='geometry', crs=DEFAULT_EPSG.value)
    elif isinstance(df, gpd.GeoDataFrame):
        pass # already a GeoDataFrame
    else:
        raise TypeError(f"Unsupported type {type(df)}. Expected pandas DataFrame, geopandas GeoDataFrame, or spark DataFrame.")
    df.to_file(path, layer='dataframe', driver="GPKG")


import pandas as pd
import pathlib

from libadalina_core.sedona_utils import to_spark_dataframe, EPSGFormats
from spatial_operators import polygonize

if __name__ == "__main__":
    """Example of reading a pandas DataFrame, converting it to a Spark DataFrame, and polygonizing its geometry."""

    # Read a CSV file into a pandas DataFrame
    df = pd.read_csv(
        pathlib.Path(__file__).parent.parent / "tests" / "samples" / "milano" / "gis_osm_roads_free_1.csv",
        sep=',', quotechar='"', dtype={'name': 'string', 'ref': 'string'},
    )
    # Fill NaN values
    df.loc[:, 'name'] = df['name'].fillna('')
    df.loc[:, 'ref'] = df['ref'].fillna('')

    # Convert the pandas DataFrame into a Spark DataFrame setting the EPSG format of the geometry
    df = to_spark_dataframe(df, EPSGFormats.EPSG4326)
    # Transform lines and points into polygons
    df = polygonize(df, 100)

    df.show()


import geopandas as gpd

from libadalina_core.sedona_utils import DEFAULT_EPSG


def geopackage_to_dataframe(path: str, layer: str) -> gpd.GeoDataFrame:
    """
    Read a GeoPackage file into a GeoDataFrame.

    Geometry is automatically converted in libadalina default EPSG `DEFAULT_EPSG`.

    :param path: The path to the GeoPackage file.
    :param layer: The layer name of the GeoPackage.
    :return: A GeoDataFrame containing the data from the specified layer.
    """
    gdf = gpd.read_file(path, layer=layer)
    gdf.to_crs(epsg=DEFAULT_EPSG.value, inplace=True)
    return gdf


from libadalina_core.readers.geopackage import geopackage_to_dataframe
import pathlib
import pandas as pd

from libadalina_core.spatial_operators.multi_table_operators import spatial_join, JoinType, spatial_aggregation, AggregationType, \
    AggregationFunction

if __name__ == "__main__":
    """Example of how to use libadalina to find the total amount of the population living in specific provinces in Italy."""

    # Set pandas display options
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', 100)

    population = geopackage_to_dataframe(
        str(pathlib.Path(__file__).parent.parent / "tests" / "samples" / "population-north-italy" / "nord-italia.gpkg"),
        "census2021"
    )[['T', 'geometry']]

    regions = geopackage_to_dataframe(
        str(pathlib.Path(__file__).parent.parent / "tests" / "samples" / "regions" / "NUTS_RG_20M_2024_4326.gpkg"),
        "NUTS_RG_20M_2024_4326.gpkg"
    )[["LEVL_CODE", "NUTS_NAME", "CNTR_CODE", "geometry"]]

    # select province of Milan and Cremona
    filtered_regions = regions[
        (regions['LEVL_CODE'] == 3) &
        (regions['CNTR_CODE'] == "IT") &
        (regions['NUTS_NAME'].str.contains('Milano|Cremona', case=False))
    ]

    # join with population table to get the population of these provinces
    result = spatial_aggregation(
        spatial_join(filtered_regions, population, join_type=JoinType.LEFT)
              # join operator renames the geometries adding suffixes _left and _right to avoid conflicts
              .withColumnRenamed('geometry_left', 'geometry'),
        aggregate_functions=[
            AggregationFunction("T", AggregationType.SUM, 'population', proportional='geometry_right'),
    ])
    result.show(truncate=False)




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('../..'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'libadalina-core'
copyright = '2025, OptLab, University of Milan'
author = 'Marco Casazza, Alberto Ceselli, Marco Premoli'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Intersphinx settings
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
    'geopandas': ('https://geopandas.org/en/stable/', None),
    'pyspark': ('https://spark.apache.org/docs/latest/api/python/', None),
}

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']

latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '10pt',
    'extraclassoptions': 'oneside',
    'passoptionstopackages': r'''
        \PassOptionsToPackage{svgnames}{xcolor}
        ''',
    'fncychap': r'\usepackage[Bjornstrup]{fncychap}',
    'makeindex': r'\usepackage[columns=1]{idxlayout}\makeindex'
}


import logging
import os
import jdk

def install_jdk_if_needed():
    """
    Install a compatible JDK if `JAVA_HOME` environment variable is not found.
    """
    if 'JAVA_HOME' not in os.environ:
        version = '17'
        jre_path = os.path.join(jdk._JRE_DIR, version)

        if os.path.exists(jre_path):
            logging.info(f'JAVA_HOME not set but JRE already downloaded')
            os.environ['JAVA_HOME'] = jre_path
        else:
            logging.info('JAVA_HOME not set, installing JRE...')
            java_home = jdk.install(version, jre=True)
            os.environ['JAVA_HOME'] = java_home
            os.symlink(java_home, jre_path)

    logging.info(f'JAVA_HOME set to {os.environ.get("JAVA_HOME")}')


from enum import Enum

class EPSGFormats(Enum):
    """
    Enum representing common EPSG formats used in geospatial data.
    """
    EPSG4326 = 4326  # WGS84
    EPSG32632 = 32632  # UTM zone 32N

    @staticmethod
    def from_code(code: int) -> 'EPSGFormats':
        for f in EPSGFormats:
            if f.value == code:
                return f
        raise ValueError(f"No EPSG format found for code {code}")

"""
Default EPSG format used in libadalina.

All DataFrame are converted upon reading and writing to this format.
"""
DEFAULT_EPSG = EPSGFormats.EPSG4326


from enum import Enum
from dataclasses import dataclass
from .commons import DataFrame
from sedona.sql import ST_Area, ST_Intersection, ST_Union, ST_Buffer, ST_GeometryType, ST_Dump
from libadalina_core.sedona_utils import to_spark_dataframe, DEFAULT_EPSG
import pyspark.sql.functions as func
import pyspark.sql as ps

def polygonize(df: DataFrame, radius_meters: float) -> ps.DataFrame:
    """
    Transform lines and points into polygons by buffering them with a given radius.

    Each line (or multi-line) is transformed into a polygon by buffering it on both sides,
    while points are buffered to create a circular area around them.

    Geometries are implicitly converted to :py:DEFAULT_EPSG.

    :param df: The input DataFrame containing geometries.
    :param radius_meters: The radius in meters to use for buffering points and lines.
    :return: A Spark DataFrame with a new column 'polygonized_geometry' containing the buffered geometries.

    """
    table = to_spark_dataframe(df)
    return table.select("*", func
                        .when(ST_GeometryType(table.geometry).like('%Point%'),
                              ST_Buffer(func.col('geometry'), radius_meters, func.lit(True)))
                        .when(ST_GeometryType(func.col('geometry')).like('%LineString%'),
                              ST_Union(
                                  ST_Buffer(func.col('geometry'), radius_meters, func.lit(True),
                                            parameters=func.lit('endcap=flat side=left')),
                                  ST_Buffer(func.col('geometry'), radius_meters, func.lit(True),
                                            parameters=func.lit('endcap=flat side=right'))
                              ))
                        .otherwise(table.geometry)
                        .alias('polygonized_geometry')
                        )


def explode_multi_geometry(df: DataFrame) -> ps.DataFrame:
    table = to_spark_dataframe(df)

    return table.select("*", func
                        .when(df.geometry.isNull(), func.array())
                        .when(ST_GeometryType(df.geometry).like('%Multi%'),
                              func.explode(ST_Dump(df.geometry)))
                        .otherwise(df.geometry)
                        )


class AggregationType(Enum):
    COUNT = 'count'
    SUM = 'sum'
    AVG = 'avg'
    MIN = 'min'
    MAX = 'max'

    def to_spark_func(self):
        if self == AggregationType.COUNT:
            return func.count
        elif self == AggregationType.SUM:
            return func.sum
        elif self == AggregationType.AVG:
            return func.avg
        elif self == AggregationType.MIN:
            return func.min
        elif self == AggregationType.MAX:
            return func.max
        return func.count  # Default to count if none matched

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

@dataclass
class AggregationFunction:
    column: str
    aggregation_type: AggregationType
    alias: str | None = None
    proportional: str | None = None

def spatial_aggregation(table: DataFrame, aggregate_functions: list[AggregationFunction]) -> ps.DataFrame:
    table = to_spark_dataframe(table)

    columns_to_aggregate = [c.column for c in aggregate_functions]
    projection_of_not_aggregated_columns = (
        func.first(c).alias(c) for c in table.columns if c != 'geometry' and c not in columns_to_aggregate
    )

    columns_with_no_proportional_aggregation = [c for c in aggregate_functions if c.proportional is None]
    columns_with_proportional_aggregation = [c for c in aggregate_functions if c.proportional is not None]

    projection_of_aggregated_columns = (
        agg_func.aggregation_type.to_spark_func()(func.col(agg_func.column)).alias(
            f"{agg_func.aggregation_type.value}({agg_func.column})" if agg_func.alias is None else agg_func.alias
        ) for agg_func in columns_with_no_proportional_aggregation if agg_func.column in table.columns
    )

    projection_of_proportional_aggregated_columns = (
        agg_func.aggregation_type.to_spark_func()(func.col(agg_func.column) * ST_Area(ST_Intersection(func.col('geometry'), func.col(agg_func.proportional))) / ST_Area(func.col(agg_func.proportional))).alias(
            f"{agg_func.aggregation_type.value}({agg_func.column})" if agg_func.alias is None else agg_func.alias
        ) for agg_func in columns_with_proportional_aggregation if agg_func.column in table.columns
    )

    # Group by geometry and aggregate other columns
    aggregated = (table
                  .groupby(table.geometry)
                    .agg(
                        # from the columns for which is not specified an aggregation function, take the first value
                        *projection_of_not_aggregated_columns,
                        # apply the aggregation functions to the other columns
                        *projection_of_aggregated_columns,
                        *projection_of_proportional_aggregated_columns
                    ))

    return aggregated


pip install libadalina-core # install libadalina-core

from ameliadp_sql_toolkit import GrinsAmeliaSQLToolkit # import amelia package to read from their datasets
amelia_sql_toolkit = GrinsAmeliaSQLToolkit()
milan_osm = amelia_sql_toolkit.read_data(table_name='gis_osm_roads_free_1') # read a dataset

# select interesting columns and fill NaNs
milan_osm = milan_osm[['geometry', 'name']]
milan_osm.loc[:, 'name'] = milan_osm['name'].fillna('')

milan_osm

from libadalina_core.sedona_utils import to_spark_dataframe, EPSGFormats
from libadalina_core.spatial_operators import polygonize

# load dataframe into Apache Sedona
df = to_spark_dataframe(milan_osm, EPSGFormats.EPSG4326)
# Transform lines and points into polygons
df = polygonize(df, 100)

df
